"""
Routes for the product service.
"""

import asyncio
from difflib import Differ

from beanie import PydanticObjectId
from fastapi import APIRouter, HTTPException, Request, status
from fastapi.responses import RedirectResponse
from loguru import logger

from hipposerve.api.models.product import (
    CompleteProductRequest,
    CreateProductRequest,
    CreateProductResponse,
    ReadFilesResponse,
    ReadProductResponse,
    UpdateProductRequest,
    UpdateProductResponse,
)
from hipposerve.database import ProductMetadata
from hipposerve.service import acl, product, storage, users
from hipposerve.service.auth import AuthenticationError, requires

product_router = APIRouter(prefix="/product")

DEFAULT_USER_USER_NAME = "default_user"


@product_router.put("/new")
@requires(["hippo:admin", "hippo:write"])
async def create_product(
    request: Request,
    model: CreateProductRequest,
) -> CreateProductResponse:
    """
    Create a new product, returning the pre-signed URLs for the sources.
    """

    logger.info(
        "Create product request: {} from {}", model.name, request.user.display_name
    )

    if await product.exists(name=model.name):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT, detail="Product already exists"
        )

    item, presigned = await product.create(
        name=model.name,
        description=model.description,
        metadata=model.metadata,
        sources=model.sources,
        user_name=request.user.display_name,
        storage=request.app.storage,
        product_readers=model.product_readers,
        product_writers=model.product_writers,
        mutlipart_size=model.multipart_batch_size,
    )

    logger.info(
        "Successfully created {} pre-signed URL(s) for product upload {} (id: {}) from {}",
        sum(len(x) for x in presigned.values()),
        model.name,
        item.id,
        request.user.display_name,
    )

    return CreateProductResponse(id=item.id, upload_urls=presigned)


@product_router.get("/search/{text}")
@requires(["hippo:admin", "hippo:read"])
async def search(
    text: str,
    request: Request,
) -> list[ProductMetadata]:
    """
    Search for a product by name.
    """

    logger.info(
        "Search for product {} request from {}", text, request.user.display_name
    )

    items = await product.search_by_name(
        name=text, groups=request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Successfully found {} product(s) matching {} requested by {}",
        len(items),
        text,
        request.user.display_name,
    )

    return await asyncio.gather(*[item.to_metadata() for item in items])


@product_router.post("/{id}/complete")
@requires(["hippo:admin", "hippo:write"])
async def complete_product(
    id: PydanticObjectId,
    request: Request,
    model: CompleteProductRequest,
) -> None:
    """
    Complete a product's upload. Must be called before the sources are available.
    """
    logger.info(
        "Complete product request for {} from {}", id, request.user.display_name
    )

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )

    success = await product.complete(
        product=item,
        storage=request.app.storage,
        headers=model.headers,
        sizes=model.sizes,
    )

    if not success:
        raise HTTPException(
            status_code=status.HTTP_424_FAILED_DEPENDENCY,
            detail="Not all sources were present.",
        )

    logger.info("Successfully completed product {} (id: {})", item.name, item.id)


@product_router.get("/{id}")
@requires(["hippo:admin", "hippo:read"])
async def read_product(
    id: PydanticObjectId,
    request: Request,
) -> ReadProductResponse:
    """
    Read a single product's metadata.
    """

    logger.info("Read product request for {} from {}", id, request.user.display_name)

    try:
        product_item = await product.read_by_id(
            id, request.user.groups, scopes=request.auth.scopes
        )
        item = await product_item.to_metadata()

        response = ReadProductResponse(
            current_present=item.current,
            current=item.version if item.current else None,
            requested=item.version,
            versions={item.version: item},
        )

        logger.info(
            "Successfully read product {} (id: {}) requested by {}",
            item.name,
            item.id,
            request.user.display_name,
        )

        return response
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )


@product_router.get("/{id}/tree")
@requires(["hippo:admin", "hippo:read"])
async def read_tree(
    id: PydanticObjectId,
    request: Request,
) -> ReadProductResponse:
    """
    Read a single product's entire history.
    """

    logger.info(
        "Read product tree request for {} from {}", id, request.user.display_name
    )

    try:
        requested_item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
        current_item = await product.walk_to_current(
            product=requested_item,
            groups=request.user.groups,
            scopes=request.auth.scopes,
        )
        history = await product.walk_history(product=current_item)

        if not current_item.current:
            raise HTTPException(
                status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail="Unable to find the current version of the requested item",
            )

        response = ReadProductResponse(
            current_present=True,
            current=current_item.version,
            requested=requested_item.version,
            versions=history,
        )

        logger.info(
            "Successfully read product tree for {} (id: {}) requested by {}",
            requested_item.name,
            requested_item.id,
            request.user.display_name,
        )

        return response
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )


@product_router.get("/{id}/files")
@requires(["hippo:admin", "hippo:read"])
async def read_files(id: PydanticObjectId, request: Request) -> ReadFilesResponse:
    """
    Read a single product's including pre-signed URLs for downloads.
    """

    logger.info("Read files request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )

    files = await product.read_files(product=item, storage=request.app.storage)

    logger.info(
        "Read {} pre-signed URLs for product {} (id: {}) requested by {}",
        len(files),
        item.name,
        item.id,
        request.user.display_name,
    )

    return ReadFilesResponse(
        product=await item.to_metadata(),
        files=files,
    )


@product_router.get("/{id}/{slug}")
@requires(["hippo:admin", "hippo:read"])
async def read_slug(request: Request, id: str, slug: str) -> RedirectResponse:
    NOT_FOUND = HTTPException(
        status_code=status.HTTP_404_NOT_FOUND,
        detail=f"Slug {slug} not found for product {id}",
    )
    try:
        product_instance = await product.read_by_id(
            id, request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise NOT_FOUND

    # If they have not 401/403'd from reading the product, they can read sources
    try:
        presigned = await storage.read(
            file=product_instance.sources[slug], storage=request.app.storage
        )
    except (FileNotFoundError, KeyError):
        raise NOT_FOUND

    return RedirectResponse(url=presigned, status_code=status.HTTP_302_FOUND)


@product_router.post("/{id}/diff")
@requires(["hippo:admin", "hippo:write"])
async def metadata_diff(
    id: PydanticObjectId,
    model: UpdateProductRequest,
    request: Request,
):
    """
    Calculates a diff between the current product and any new metadata.
    """

    logger.info("Update product request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )

    try:
        acl.check_user_access(
            user_groups=request.user.groups,
            document_groups=item.writers,
            scopes=request.auth.scopes,
        )
    except AuthenticationError:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="User does not have write access to this product",
        )

    # Check the _new_ owner is an existing user
    if model.owner is not None:
        try:
            await users.confirm_user(name=model.owner)
        except users.UserNotFound:
            raise HTTPException(
                status_code=status.HTTP_406_NOT_ACCEPTABLE, detail="User not found"
            )

    if not item.current:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Cannot update a product that is not the current version",
        )

    diff = await product.metadata_diff(
        product=item,
        name=model.name,
        description=model.description,
        metadata=model.metadata,
        level=model.level,
    )

    if diff.get("description", None) is not None:
        differ = Differ()
        diffed_description = list(
            differ.compare(
                item.description.splitlines(),
                model.description.splitlines() if model.description else [],
            )
        )

        output_diff = "<table class='table table-bordered'><tbody>\n"
        for line, next_line in zip(diffed_description, diffed_description[1:] + [""]):
            if line.startswith("? "):
                continue
            if line.startswith("+ "):
                line = line[2:]
                if next_line.startswith("? "):
                    # Two lines: first is addition, second is location in the original line
                    position = next_line[2:]

                    new_line = ""
                    in_diff = False

                    for outchar, indicator in zip(line, position + " " * len(line)):
                        if indicator != " ":
                            if not in_diff:
                                new_line += "<b class='text-success'>"
                                in_diff = True
                                new_line += outchar
                                continue
                            else:
                                new_line += outchar
                                continue

                        if in_diff:
                            new_line += "</b>"
                            in_diff = False

                        new_line += outchar

                    line = new_line
                output_diff += f"<tr class='table-success'><td><b class='text-success'>+</b></td><td class='text-success'>{line}</td></tr>\n"
            elif line.startswith("- "):
                line = line[2:]
                if next_line.startswith("? "):
                    # Two lines: first is addition, second is location in the original line
                    position = next_line[2:]

                    new_line = ""
                    in_diff = False

                    for outchar, indicator in zip(line, position + " " * len(line)):
                        if indicator != " ":
                            if not in_diff:
                                new_line += "<b class='text-danger'>"
                                in_diff = True
                                new_line += outchar
                                continue
                            else:
                                new_line += outchar
                                continue

                        if in_diff:
                            new_line += "</b>"
                            in_diff = False

                        new_line += outchar

                    line = new_line
                output_diff += f"<tr class='table-danger'><td><b class='text-danger'>-</b></td><td class='text-danger'>{line}</td></tr>\n"
            else:
                output_diff += f"<tr><td></td><td>{line[2:]}</td></tr>\n"

        output_diff += "</tbody></table>\n"
        diff["description"] = output_diff

    logger.debug(
        "Found metadata diff for product {} (id: {}): {}", item.name, item.id, diff
    )

    return diff


@product_router.post("/{id}/update")
@requires(["hippo:admin", "hippo:write"])
async def update_product(
    id: PydanticObjectId,
    model: UpdateProductRequest,
    request: Request,
) -> UpdateProductResponse:
    """
    Update a product's details.
    """

    logger.info("Update product request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )

    try:
        acl.check_user_access(
            user_groups=request.user.groups,
            document_groups=item.writers,
            scopes=request.auth.scopes,
        )
    except AuthenticationError:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="User does not have write access to this product",
        )

    # Check the _new_ owner is an existing user
    if model.owner is not None:
        try:
            await users.confirm_user(name=model.owner)
        except users.UserNotFound:
            raise HTTPException(
                status_code=status.HTTP_406_NOT_ACCEPTABLE, detail="User not found."
            )

    if model.level:
        try:
            new_product, upload_urls = await product.update(
                product=item,
                access_groups=request.user.groups,
                scopes=request.auth.scopes,
                name=model.name,
                description=model.description,
                metadata=model.metadata,
                new_sources=model.new_sources,
                replace_sources=model.replace_sources,
                drop_sources=model.drop_sources,
                storage=request.app.storage,
                level=model.level,
            )
        except AuthenticationError:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail="User does not have write access to this product",
            )

        logger.info(
            "Successfully updated product {} (new id: {}; {}; old id: {}; {}) from {}",
            new_product.name,
            new_product.id,
            new_product.version,
            item.id,
            item.version,
            request.user.display_name,
        )
    else:
        new_product = item
        upload_urls = {}

    new_product = await acl.update_access_control(
        doc=new_product,
        owner=model.owner,
        add_readers=model.add_readers,
        remove_readers=model.remove_readers,
        add_writers=model.add_writers,
        remove_writers=model.remove_writers,
    )

    logger.info(
        "Successfully updated product {} (id: {}) with new access control policy",
        new_product.name,
        new_product.id,
    )

    return UpdateProductResponse(
        version=new_product.version,
        id=new_product.id,
        upload_urls=upload_urls,
    )


@product_router.post("/{id}/confirm")
@requires(["hippo:admin", "hippo:read"])
async def confirm_product(id: PydanticObjectId, request: Request) -> None:
    """
    Confirm a product's sources.
    """

    logger.info("Confirm product request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
        success = await product.confirm(
            product=item,
            storage=request.app.storage,
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )

    if not success:
        raise HTTPException(
            status_code=status.HTTP_424_FAILED_DEPENDENCY,
            detail="Not all sources were present.",
        )

    logger.info("Successfully confirmed product {} (id: {})", item.name, item.id)


@product_router.post("/{id}/pin")
@requires(["hippo:admin"])
async def pin_product(
    id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Pin a product to place it on the homepage.
    """

    logger.info("Pin product request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )

    item.pinned = True
    await item.save()

    logger.info("Successfully pinned product {} (id: {})", item.name, item.id)


@product_router.post("/{id}/unpin")
@requires(["hippo:admin"])
async def unpin_product(
    id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Unpin a product to remvove it from the homepage.
    """

    logger.info("Unpin product request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )

    item.pinned = False
    await item.save()

    logger.info("Successfully unpinned product {} (id: {})", item.name, item.id)


@product_router.delete("/{id}")
@requires(["hippo:admin", "hippo:write"])
async def delete_product(
    id: PydanticObjectId,
    request: Request,
    data: bool = False,
) -> None:
    """
    Delete a product.
    """

    logger.info(
        "Delete (single) product request for {} from {}", id, request.user.display_name
    )

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )

    await product.delete_one(
        product=item,
        access_groups=request.user.groups,
        storage=request.app.storage,
        data=data,
        scopes=request.auth.scopes,
    )

    logger.info(
        "Successfully deleted product {} (id: {}) including data"
        if data
        else "Successfully deleted product {} (id: {}) excluding data",
        item.name,
        item.id,
    )


@product_router.delete("/{id}/tree")
@requires(["hippo:admin", "hippo:write"])
async def delete_tree(
    id: PydanticObjectId,
    request: Request,
    data: bool = False,
) -> None:
    """
    Delete a product.
    """

    logger.info(
        "Delete (tree) product request for {} from {}", id, request.user.display_name
    )

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )

    await product.delete_tree(
        product=item,
        access_groups=request.user.groups,
        storage=request.app.storage,
        data=data,
        scopes=request.auth.scopes,
    )

    logger.info(
        "Successfully deleted product tree for {} (id: {}) including data"
        if data
        else "Successfully deleted product tree for {} (id: {}) excluding data",
        item.name,
        item.id,
    )



class PreflightFailedError(Exception):
    pass


class InvalidSlugError(KeyError):
    pass


class ProductIncompleteError(Exception):
    pass


class CollectionIncompleteError(Exception):
    pass



"""
Pydantic models for the product API layer.
"""

from beanie import PydanticObjectId
from pydantic import BaseModel

from hippometa import ALL_METADATA_TYPE
from hipposerve.service.product import PostUploadFile, PreUploadFile, ProductMetadata
from hipposerve.service.versioning import VersionRevision


class CreateProductRequest(BaseModel):
    name: str
    description: str
    metadata: ALL_METADATA_TYPE
    sources: dict[str, PreUploadFile]
    product_readers: list[str] = []
    product_writers: list[str] = []
    multipart_batch_size: int = 50 * 1024 * 1024


class CreateProductResponse(BaseModel):
    id: PydanticObjectId
    upload_urls: dict[str, list[str]]


class CompleteProductRequest(BaseModel):
    headers: dict[str, list[dict[str, str]]]
    sizes: dict[str, list[int]]


class ReadProductResponse(BaseModel):
    current_present: bool
    current: str | None
    requested: str
    versions: dict[str, ProductMetadata]


class ReadFilesResponse(BaseModel):
    product: ProductMetadata
    files: dict[str, PostUploadFile]


class UpdateProductRequest(BaseModel):
    name: str | None = None
    description: str | None = None
    metadata: ALL_METADATA_TYPE | None = None
    owner: str | None = None
    new_sources: dict[str, PreUploadFile] = {}
    replace_sources: dict[str, PreUploadFile] = {}
    drop_sources: list[str] = []
    level: VersionRevision | None = VersionRevision.MINOR
    add_readers: list[str] = []
    remove_readers: list[str] = []
    add_writers: list[str] = []
    remove_writers: list[str] = []


class UpdateProductResponse(BaseModel):
    version: str
    id: PydanticObjectId
    upload_urls: dict[str, list[str]]



"""
Methods for interacting with the product layer of the hippo API.
"""

from pathlib import Path

from httpx import Client, Response
from rich.console import Console
from tqdm import tqdm

from hippoclient.downloads import downloader, file_info
from hippometa import ALL_METADATA_TYPE
from hippometa.simple import SimpleMetadata
from hipposerve.api.models.product import ReadProductResponse
from hipposerve.database import ProductMetadata
from hipposerve.service.product import PostUploadFile, PreUploadFile

from .core import MultiCache
from .tools import slugify as apply_slugify

MULTIPART_UPLOAD_SIZE = 50 * 1024 * 1024


def __upload_sources(
    initial_response: Response,
    client: Client,
    sources: dict[str, PreUploadFile],
    this_product_id: str,
    console: Console | None = None,
) -> dict[str, list[dict[str, str]]]:
    responses = {}
    sizes = {}

    # Upload the sources to the presigned URLs.
    for source in sources.values():
        with source.open("rb") as file:
            if console:
                console.print("Uploading file:", source.name)

            upload_urls = initial_response.json()["upload_urls"][source.name]
            headers = []
            size = []

            # We need to handle our own redirects because otherwise the head of the file will be incorrect,
            # and we will end up with Content-Length errors.

            with tqdm(
                desc="Uploading",
                total=source.stat().st_size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as t:
                file_position = 0

                for upload_url in upload_urls:
                    retry = True

                    while retry:
                        file.seek(file_position)

                        data = file.read(MULTIPART_UPLOAD_SIZE)

                        individual_response = client.put(
                            upload_url.strip(),
                            content=data,
                            follow_redirects=True,
                            auth=None,
                            # Blocks are 50 MB so may timeout on slow connections
                            # (httpx defaults to 5 seconds)
                            timeout=120.0,
                        )

                        if individual_response.status_code in [301, 302, 307, 308]:
                            if console:
                                console.print(
                                    f"Redirected to {individual_response.headers['Location']} from {upload_url}"
                                )
                            upload_url = individual_response.headers["Location"]

                            continue
                        else:
                            retry = False
                            individual_response.raise_for_status()
                            break

                    headers.append(dict(individual_response.headers))
                    size.append(len(data))

                    file_position += MULTIPART_UPLOAD_SIZE
                    t.update(len(data))

            responses[source.name] = headers
            sizes[source.name] = size

            if console:
                console.print("Successfully uploaded file:", source.name)

    if sources:
        # Close out the upload.
        response = client.post(
            f"/product/{this_product_id}/complete",
            json={"headers": responses, "sizes": sizes},
        )

        response.raise_for_status()

        return response.json()


def create(
    client: Client,
    name: str,
    description: str,
    metadata: ALL_METADATA_TYPE,
    sources: dict[str, Path],
    source_descriptions: dict[str, str | None],
    readers: list[str] | None = None,
    writers: list[str] | None = None,
    console: Console | None = None,
) -> str:
    """
    Create a product in hippo.

    This is a multi-stage process, where we:

    a) Check and validate the sources.
    b) Make a request to hippo to create the product.
    c) Upload the sources to the presigned URLs.
    d) Confirm the upload to hippo.

    Arguments
    ---------
    client: Client
        The client to use for interacting with the hippo API.
    name : str
        The name of the product.
    description : str
        The description of the product.
    metadata : ALL_METADATA_TYPE
        The metadata of the product, as a validated pydantic model.
    sources : dict[str, Path]
        The list of paths to the sources of the product keyed by their slug.
    source_descriptions : dict[str, str | None]
        The descriptions of the sources keyed by their slug.
    readers : list[str] | None, optional
        A list of groups that can read the product.
    writers : list[str] | None, optional
        A list of groups that can write to the product.
    console : Console, optional
        The rich console to print to.

    Returns
    -------
    str
        The ID of the product created.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    # Check and validate the sources.
    assert len(sources) == len(source_descriptions)

    # Co-erce sources to paths as they will inevitably be strings...
    sources = {x: Path(y) for x, y in sources.items()}

    source_metadata = {}

    for slug in sources:
        filename = sources[slug]
        desc = source_descriptions[slug]

        source_metadata[slug] = file_info(filename, desc)

        if console:
            console.print("Successfully validated file:", source_metadata[slug])

    # Make a request to hippo to create the product.
    if metadata is None:
        metadata = SimpleMetadata()

    content = {
        "name": name,
        "description": description,
        "metadata": metadata.model_dump(mode="json"),
        "sources": source_metadata,
        "mutlipart_batch_size": MULTIPART_UPLOAD_SIZE,
    }

    if readers is not None:
        content["product_readers"] = readers
    if writers is not None:
        content["product_writers"] = writers

    response = client.put("/product/new", json=content)

    response.raise_for_status()

    this_product_id = response.json()["id"]

    __upload_sources(
        initial_response=response,
        client=client,
        sources=sources,
        this_product_id=this_product_id,
        console=console,
    )

    if console:
        console.print(
            f"Successfully created product {this_product_id} in remote database."
        )

    # Confirm the upload to hippo.
    response = client.post(f"/product/{this_product_id}/confirm")

    response.raise_for_status()

    if console:
        console.print(f"Successfully completed upload of {name}.", style="bold green")

    return this_product_id


def read_with_versions(
    client: Client, id: str, console: Console | None = None
) -> ReadProductResponse:
    """
    Read a product from hippo by ID. Always returns the requested version,
    without information about other versions.

    Arguments
    ---------
    client: Client
        The client to use for interacting with the hippo API.
    id : str
        The ID of the product to read.
    console : Console, optional
        The rich console to print to.

    Returns
    -------
    ReadProductResponse
        The response from the API.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.get(f"/product/{id}")

    response.raise_for_status()

    model = ReadProductResponse.model_validate_json(response.content)

    if console:
        console.print(f"Successfully read product {id}")

    return model


def read(client: Client, id: str, console: Console | None = None) -> ProductMetadata:
    """
    Read a product from hippo by ID. Always returns the requested version,
    without information about other versions.

    Arguments
    ---------
    client: Client
        The client to use for interacting with the hippo API.
    id : str
        The ID of the product to read.
    console : Console, optional
        The rich console to print to.

    Returns
    -------
    ReadProductResponse
        The response from the API.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    model = read_with_versions(client, id)

    # This model includes version history. We don't want that.
    model = model.versions[model.requested]

    if console:
        console.print(f"Successfully read product ({model.name})")

    return model


def update(
    client: Client,
    id: str,
    name: str | None,
    description: str | None,
    level: int,
    metadata: ALL_METADATA_TYPE | None,
    new_sources: dict[str, Path] | None = None,
    new_source_descriptions: dict[str, str | None] | None = None,
    replace_sources: dict[str, Path] | None = None,
    replace_source_descriptions: dict[str, str | None] | None = None,
    drop_sources: list[str] | None = None,
    add_readers: list[str] | None = None,
    remove_readers: list[str] | None = None,
    add_writers: list[str] | None = None,
    remove_writers: list[str] | None = None,
    console: Console | None = None,
) -> str:
    """
    Update a product in hippo.

    Arguments
    ----------
    client: Client
        The client to use for interacting with the hippo API.
    id : str
        The ID of the product to update.
    name : str | None
        The new name of the product.
    description : str | None
        The new description of the product.
    level : int
        The new version level where 0 is major, 1 is minor, and 2 is patch.
    metadata : ALL_METADATA_TYPE
        The new product metadata to update.
    new_sources : dict[str, Path] | None
        A list of new sources to add to the product.
    new_source_descriptions : dict[str, str | None] | None
        A list of descriptions for the new sources keyed by their slug.
    replace_sources : dict[str, Path] | None
        A list of sources to replace in a product.
    replace_source_descriptions : dict[str, str | None] | None
        A list of descriptions for the sources to replace keyed by their slug.
    drop_sources : list[str]
        A list of source IDs to delete from a product.
    console : Console, optional
        The rich console to print to.

    Returns
    -------
    str
        The ID of the updated product.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    new_source_metadata = {}

    for slug in (new_sources or {}).keys():
        filename = new_sources[slug]
        desc = new_source_descriptions[slug]

        new_source_metadata[slug] = file_info(filename, desc)

        if console:
            console.print("Successfully validated file:", new_source_metadata[slug])

    replace_source_metadata = {}
    for slug in (replace_sources or {}).keys():
        filename = replace_sources[slug]
        desc = replace_source_descriptions[slug]

        replace_source_metadata[slug] = file_info(filename, desc)

        if console:
            console.print("Successfully validated file:", replace_source_metadata[slug])

    response = client.post(
        f"/product/{id}/update",
        json={
            "name": name,
            "description": description,
            "level": level,
            "metadata": metadata.model_dump() if metadata else None,
            "new_sources": new_source_metadata,
            "replace_sources": replace_source_metadata,
            "drop_sources": drop_sources or [],
            "add_readers": add_readers or [],
            "remove_readers": remove_readers or [],
            "add_writers": add_writers or [],
            "remove_writers": remove_writers or [],
        },
    )

    response.raise_for_status()

    this_product_id = response.json()["id"]

    __upload_sources(
        initial_response=response,
        client=client,
        sources={**(new_sources or {}), **(replace_sources or {})},
        this_product_id=this_product_id,
        console=console,
    )

    if console:
        console.print(f"Successfully updated product {id}.", style="bold green")

    return this_product_id


def delete(client: Client, id: str, console: Console | None = None) -> bool:
    """
    Delete a product from hippo.

    Arguments
    ----------
    client: Client
        The client to use for interacting with the hippo API.
    id : str
        The ID of the product to delete.
    console : Console, optional
        The rich console to print to.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.delete(f"/product/{id}")

    response.raise_for_status()

    if console:
        console.print(f"Successfully deleted product {id}.", style="bold green")

    return True


def search(
    client: Client, text: str, console: Console | None = None
) -> list[ProductMetadata]:
    """
    Search for text information in products (primarily names).

    Arguments
    ----------
    client: Client
        The client to use for interacting with the hippo API.
    text : str
        The text to search for.
    console : Console, optional
        The rich console to print to.

    Returns
    -------
    list[ProductMetadata]
        The list of products that match the search query.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.get(f"/product/search/{text}")

    response.raise_for_status()

    models = [ProductMetadata.model_validate(x) for x in response.json()]

    if console:
        console.print(f"Successfully searched for products matching {text}")

    return models


def cache(
    client: Client, cache: MultiCache, id: str, console: Console | None = None
) -> list[Path]:
    """
    Cache a product from hippo.

    Arguments
    ----------
    client: Client
        The client to use for interacting with the hippo API.
    cache: MultiCache
        The cache to use for storing the product.
    id : str
        The ID of the product to cache.
    console : Console, optional
        The rich console to print to.

    Returns
    -------
    list[Path]
        The list of paths to the cached sources.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    CacheNotWriteableError
        If the cache is not writeable
    """

    response = client.get(f"/product/{id}/files")

    response.raise_for_status()

    post_upload_files = {
        x: PostUploadFile.model_validate(y) for x, y in response.json()["files"].items()
    }

    if console:
        console.print(f"Successfully read product {id}")

    response_paths = []

    for file in post_upload_files.values():
        # See if it's already cached.
        try:
            cached = cache.available(file.uuid)
            response_paths.append(cached)

            if console:
                console.print(f"Found cached file {file.name}", style="green")

            continue
        except FileNotFoundError:
            if console:
                console.print(
                    f"File {file.name} ({file.uuid}) not found in cache", style="red"
                )
            cached = None

        cached = cache.get(
            id=file.uuid,
            path=file.object_name,
            checksum=file.checksum,
            size=file.size,
            presigned_url=file.url,
        )

        response_paths.append(cached)

        if console:
            console.print(f"Cached file {file.name} ({file.uuid})", style="yellow")

    return response_paths


def download(
    client: Client,
    id: str,
    directory: Path,
    console: Console | None = None,
    slugs: list[str] | None = None,
    slugify: bool = False,
) -> Path:
    """
    Download a product from HIPPO onto the local filesystem, storing the data
    in the provided 'directory'. Data is stored as follows:

    directory/Product Name
        product.json < Metadata for this product
        slug/
          file.fits
        data/
          file2.fits

    So inside the provided directory, we create a new directory with the product's
    name. In there, each slug is stored in its own directory.

    Arguments
    ----------
    client: Client
        The client to use for interacting with the hippo API.
    id : str
        The ID of the product to download.
    console : Console, optional
        The rich console to print to.
    slugs : list[str], optional
        Only download specific slugs.
    directory: Path
        The directory to download the product into.
    slugify : bool, optional
        Whether to convert product names into 'slugified' versions
        that are esaier to read as directory names.

    Returns
    -------
    Path
        The path to the cached sources.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    FileNotFoundError
        If the directory does not exist or is not a directory.
    """

    response = client.get(f"/product/{id}/files")

    response.raise_for_status()

    post_upload_files = {
        x: PostUploadFile.model_validate(y) for x, y in response.json()["files"].items()
    }

    metadata = ProductMetadata.model_validate(response.json()["product"])

    if console:
        console.print(f"Successfully read product {id}")

    if slugify:
        metadata.name = apply_slugify(metadata.name)

    product_directory = directory / metadata.name

    # Raises FileNotFoundError if `directory` does not exist.
    product_directory.mkdir(exist_ok=True, parents=False)

    with open(product_directory / "product.json", "w") as handle:
        handle.write(metadata.model_dump_json(indent=2))

        if console:
            console.print(f"Successfully wrote metadata for product {metadata.name}")

    for source_slug, source_data in post_upload_files.items():
        if slugs and source_slug not in slugs:
            continue

        slug_directory = product_directory / source_slug
        slug_directory.mkdir(exist_ok=True)
        slug_path = slug_directory / source_data.name

        # Check if it's already there and if it matches our checksum
        if slug_path.exists():
            if file_info(slug_path)["checksum"] == source_data.checksum:
                if console:
                    console.print(f"Valid {source_slug} already exists at {slug_path}")
                continue

        downloader(
            presigned_url=source_data.url,
            output_destination=slug_path,
            console=console,
        )

        if console:
            console.print(f"Downloaded slug {source_slug} to {slug_path}")

    return product_directory


def uncache(
    client: Client, cache: MultiCache, id: str, console: Console | None = None
) -> None:
    """
    Clear the cache of a product.

    Arguments
    ----------
    client: Client
        The client to use for interacting with the hippo API.
    cache: MultiCache
        The cache to use for storing the product.
    id : str
        The ID of the product to remove from the cache.
    console : Console, optional
        The rich console to print to.
    """

    product = read(client, id)

    for source in product.sources.values():
        cache.remove(source.uuid)

        if console:
            console.print(f"Removed file {source.name} ({source.uuid}) from cache")

    return


def product_add_reader(
    client: Client, id: str, group: str, console: Console | None = None
) -> str:
    response = client.post(
        f"/product/{id}/update",
        json={
            "level": None,
            "add_readers": [group],
        },
    )

    response.raise_for_status()
    this_product_id = response.json()["id"]
    if console:
        console.print(
            f"Successfully added {group} to product {id} readers.", style="bold green"
        )

    return this_product_id


def product_remove_reader(
    client: Client, id: str, group: str, console: Console | None = None
) -> str:
    response = client.post(
        f"/product/{id}/update",
        json={
            "level": None,
            "remove_readers": [group],
        },
    )

    response.raise_for_status()
    this_product_id = response.json()["id"]
    if console:
        console.print(
            f"Successfully removed {group} from product {id} readers.",
            style="bold green",
        )

    return this_product_id


def product_add_writer(
    client: Client, id: str, group: str, console: Console | None = None
) -> str:
    response = client.post(
        f"/product/{id}/update",
        json={
            "level": None,
            "add_writers": [group],
        },
    )

    response.raise_for_status()
    this_product_id = response.json()["id"]
    if console:
        console.print(
            f"Successfully added {group} to product {id} writers.", style="bold green"
        )

    return this_product_id


def product_remove_writer(
    client: Client, id: str, group: str, console: Console | None = None
) -> str:
    response = client.post(
        f"/product/{id}/update",
        json={
            "level": None,
            "remove_writers": [group],
        },
    )

    response.raise_for_status()
    this_product_id = response.json()["id"]
    if console:
        console.print(
            f"Successfully removed {group} from product {id} writers.",
            style="bold green",
        )

    return this_product_id



from pathlib import Path
from typing import Optional

import httpx
import xxhash
from rich.console import Console
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)


def downloader(
    presigned_url: str,
    output_destination: Path,
    console: Optional[Console] = None,
):
    with httpx.Client() as client:
        with client.stream("GET", presigned_url, follow_redirects=True) as response:
            response.raise_for_status()
            total = int(response.headers.get("Content-Length", 0))

            if console:
                progress = Progress(
                    TextColumn("[progress.description]{task.description}"),
                    BarColumn(),
                    DownloadColumn(),
                    TransferSpeedColumn(),
                    TimeRemainingColumn(),
                    console=console,
                )

                task_id = progress.add_task("Downloading", total=total)

            with progress, open(output_destination, "wb") as f:
                for chunk in response.iter_bytes(chunk_size=8192):
                    f.write(chunk)
                    if console:
                        progress.update(task_id, advance=len(chunk))


def file_info(filename: Path, description: str | None = None) -> dict:
    with open(filename, "rb") as handle:
        return {
            "name": filename.name,
            "size": filename.stat().st_size,
            "checksum": f"xxh64:{xxhash.xxh64(handle.read()).hexdigest()}",
            "description": description,
        }



"""
Methods for interacting with the collections layer of the hippo API
"""

from pathlib import Path

from httpx import Client
from rich.console import Console

from hipposerve.api.models.relationships import ReadCollectionResponse

from .core import MultiCache
from .product import cache as cache_product
from .product import download as download_product
from .product import uncache as uncache_product
from .tools import slugify as apply_slugify


def create(
    client: Client,
    name: str,
    description: str,
    readers: list[str] | None = None,
    writers: list[str] | None = None,
    console: Console | None = None,
) -> str:
    """
    Create a new collection in hippo.

    Arguments
    ---------
    client : Client
        The client to use for interacting with the hippo API.
    name : str
        The name of the collection.
    description : str
        The description of the collection.
    readers : list[str], optional
        A list of groups that can read the collection.
    writers : list[str], optional
        A list of groups that can write to the collection.
    console : Console, optional
        The Console to use to print to.

    Returns
    -------
    str
        The ID of the collection created.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    content = {"description": description}

    if readers is not None:
        content["readers"] = readers
    if writers is not None:
        content["writers"] = writers

    response = client.put(f"/relationships/collection/{name}", json=content)

    response.raise_for_status()

    if console:
        console.print(f"Successfully created collection {name}.", style="bold green")

    return response.json()


def read(
    client: Client, id: str, console: Console | None = None
) -> ReadCollectionResponse:
    """
    Read a collection from hippo.

    Arguments
    ---------
    client: Client
        The client to use for interacting with the hippo API.
    id : str
        The id of the collection to read.
    console : Console, optional
        The Console to use to print to.

    Returns
    -------
    ReadCollectionResponse
        The response from the API.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.get(f"/relationships/collection/{id}")

    response.raise_for_status()

    model = ReadCollectionResponse.model_validate_json(response.content)

    if console:
        console.print(f"Successfully read collection {model.name} ({id})")

    return model


def add_reader(
    client: Client, id: str, group: str, console: Console | None = None
) -> str:
    response = client.post(
        f"/relationships/collection/{id}", json={"add_readers": [group]}
    )
    response.raise_for_status()
    if console:
        console.print(f"Successfully added {group} to collection {id} readers.")
    this_collection_id = response.json()
    return this_collection_id


def remove_reader(
    client: Client, id: str, group: str, console: Console | None = None
) -> str:
    response = client.post(
        f"/relationships/collection/{id}", json={"remove_readers": [group]}
    )
    response.raise_for_status()
    if console:
        console.print(f"Successfully removed {group} from collection {id} readers.")
    this_collection_id = response.json()
    return this_collection_id


def add_writer(
    client: Client, id: str, group: str, console: Console | None = None
) -> str:
    response = client.post(
        f"/relationships/collection/{id}", json={"add_writers": [group]}
    )
    response.raise_for_status()
    if console:
        console.print(f"Successfully added {group} to collection {id} writers.")
    this_collection_id = response.json()
    return this_collection_id


def remove_writer(
    client: Client, id: str, group: str, console: Console | None = None
) -> str:
    response = client.post(
        f"/relationships/collection/{id}", json={"remove_writers": [group]}
    )
    response.raise_for_status()
    if console:
        console.print(f"Successfully removed {group} from collection {id} writers.")
    this_collection_id = response.json()
    return this_collection_id


def search(
    client: Client, name: str, console: Console | None = None
) -> list[ReadCollectionResponse]:
    """
    Search for collections in hippo.

    Arguments
    ---------
    client: Client
        The client to use for interacting with the hippo API.
    name : str
        The name of the collection to search for.
    console: Console, optional
        The rich console to print to.

    Returns
    -------
    list[ReadCollectionResponse]
        The response from the API.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.get(f"/relationships/collection/search/{name}")

    response.raise_for_status()

    models = [ReadCollectionResponse.model_validate(x) for x in response.json()]

    if console:
        console.print(f"Successfully searched for collection {name}")

    return models


def add(client: Client, id: str, product: str, console: Console | None = None) -> bool:
    """
    Add a product to a collection in hippo.

    Arguments
    ---------
    client: Client
        The client to use for interacting with the hippo API.
    id : str
        The id of the collection to add the product to.
    product : str
        The id of the product to add to the collection.
    console: Console, optional
        The rich console to print to.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.put(f"/relationships/collection/{id}/{product}")

    response.raise_for_status()

    if console:
        console.print(
            f"Successfully added product {product} to collection {id}.",
            style="bold green",
        )

    return True


def remove(
    client: Client, id: str, product: str, console: Console | None = None
) -> bool:
    """
    Remove a product from a collection in hippo.

    Arguments
    ---------
    client: Client
        The client to use for interacting with the hippo API.
    name : str
        The id of the collection to remove the product from.
    product : str
        The id of the product to remove from the collection.
    console: Console, optional
        The rich console to print to.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.delete(f"/relationships/collection/{id}/{product}")

    response.raise_for_status()

    if console:
        console.print(
            f"Successfully removed product {product} from collection {id}.",
            style="bold green",
        )

    return True


def delete(client: Client, id: str, console: Console | None = None) -> bool:
    """
    Delete a collection from hippo.

    Arguments
    ----------
    client: Client
        The client to use for interacting with the hippo API.
    id : str
        The name of the collection to delete.
    console: Console, optional
        The rich console to print to.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.delete(f"/relationships/collection/{id}")

    response.raise_for_status()

    if console:
        console.print(f"Successfully deleted collection {id}.", style="bold green")

    return True


def cache(
    client: Client, cache: MultiCache, id: str, console: Console | None = None
) -> list[Path]:
    """
    Cache a collection from hippo.

    Arguments
    ---------
    client: Client
        The client to use for interacting with the hippo API.
    cache : MultiCache
        The cache to use for storing the collection.
    id : str
        The id of the collection to cache.
    console: Console, optional
        The rich console to print to.

    Returns
    -------
    list[Path]
        The paths to the cached files.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    CacheNotWriteableError
        If the cache is not writeable
    """

    collection = read(client, id)

    paths = []

    for product in collection.products:
        paths += cache_product(client, cache, product.id)

    for child in collection.child_collections:
        paths += cache(client, cache, child.id)

    return paths


def download(
    client: Client,
    id: str,
    directory: Path,
    console: Console | None = None,
    slugify: bool = False,
) -> Path:
    """
    Download a collection from HIPPO onto the local filesystem, storing the data
    in the provided 'directory'. Data is stored as follows:

    directory/Collection Name
      collection.json < Metadata for this collection
      Product Name/
        ... < See product.download
      Child Collection/
        collection.json
        Product Name/
          ... < See product.download

    Inside the provided directory, we create a new directory with the collection name.
    In there we store each product and child collection recursively, and store
    the metadata for the collection as `collection.json`.

    Arguments
    ---------
    client: Client
        The client to use for interacting with the hippo API
    id: str
        The ID of the collection to download.
    directory: Path
        The directory to download the collection into.
    console : Console, optional
        The rich console to print to.
    slugify : bool, optional
        Whether to convert names to valid 'slugs'.

    Returns
    -------
    Path
        The path to the cached collection.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    FileNotFoundError
        If the directory does not exist or is not a directory.
    """

    collection = read(client, id)

    if slugify:
        collection.name = apply_slugify(collection.name)

    collection_directory = directory / collection.name

    collection_directory.mkdir(exist_ok=True, parents=False)

    with open(collection_directory / "collection.json", "w") as handle:
        handle.write(collection.model_dump_json(indent=2))

        if console:
            console.print(
                f"Successfully wrote metadata for collection {collection.name}"
            )

    for product in collection.products:
        download_product(
            client=client,
            id=product.id,
            directory=collection_directory,
            console=console,
            slugify=slugify,
        )

    for child in collection.child_collections:
        download(
            client=client,
            id=child.id,
            directory=collection_directory,
            console=console,
            slugify=slugify,
        )

    return collection_directory


def uncache(
    client: Client, cache: MultiCache, id: str, console: Console | None = None
) -> None:
    """
    Remove a collection from local caches.

    Arguments
    ---------
    client: Client
        The client to use for interacting with the hippo API.
    cache : MultiCache
        The cache to use for storing the collection.
    id : str
        The id of the collection to uncache.
    console: Console, optional
        The rich console to print to.

    Raises
    ------
    CacheNotWriteableError
        If the cache is not writeable
    """

    collection = read(client, id)

    for product in collection.products:
        uncache_product(client, cache, product.id)



"""
Authentication code for the API, which uses api key based
authentication.
"""

import functools
import inspect

from fastapi import HTTPException, Request, status
from starlette._utils import is_async_callable

AVAILABLE_GRANTS = {
    "hippo:admin",
    "hippo:read",
    "hippo:write",
}


class AuthenticationError(Exception):
    pass


def has_required_scope(request: Request, scopes: list[str]) -> bool:
    allowed = set(scopes)
    if any(scope in allowed for scope in request.auth.scopes):
        return True

    raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)


def requires(scopes: str | list[str]):
    """
    Check whether your user has the required scope.
    """

    def decorator(func):
        scopes_list = [scopes] if isinstance(scopes, str) else list(scopes)

        for scope in scopes_list:
            if scope not in AVAILABLE_GRANTS:
                raise HTTPException(501, "Invalid scope, server error")

        sig = inspect.signature(func)
        for idx, parameter in enumerate(sig.parameters.values()):
            if parameter.name == "request":
                break
        else:
            raise Exception(f'No "request" argument on function "{func}"')

        if is_async_callable(func):

            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                request = kwargs.get("request", args[idx] if idx < len(args) else None)
                assert isinstance(request, Request)

                if has_required_scope(request, scopes_list):
                    return await func(*args, **kwargs)

            return async_wrapper
        else:

            @functools.wraps(func)
            def sync_wrapper(*args, **kwargs):
                request = kwargs.get("request", args[idx] if idx < len(args) else None)
                assert isinstance(request, Request)

                if has_required_scope(request, scopes_list):
                    return func(*args, **kwargs)

            return sync_wrapper

    return decorator
